/* decode.c  ****************************************************************
 
 Twiddler Driver (userspace)
 -- Chord decoder/Event handler
 
 Copyright (C) 2005, Chris Collins <chris@collins.id.au>
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author, Christopher Collins, nor any of his
    aliases may be used to endorse or promote products derived
    from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY CHRISTOPHER COLLINS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL CHRISTOPHER COLLINS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "twiddled.h"

#define	CONFDELIM	" \t\n"

int	emitkey = 0;
unsigned char	keystate[TR_LAST];

enum dec_action_e {
	ACT_SCANCODE,
	ACT_MACRO,
	ACT_ALTTABLE,
	ACT_MOUSE
};

enum dec_macroaction_e {
	MAC_SCANCODE,
	MAC_KEYDOWN,
	MAC_KEYUP
};

struct dec_macro_s {
	enum dec_macroaction_e	action;
	unsigned short			data;
	struct dec_macro_s *	next;
};

struct dec_actiontable_s;

struct dec_action_s {
	unsigned char keystate[4];
	enum dec_action_e	action;
	union {
		unsigned short	scancode;
		struct dec_macro_s *macrostart;
		struct dec_actiontable_s *newtable;
	};
	struct dec_action_s	*next;
};

struct dec_actiontable_s {
	struct dec_action_s	*first;
	struct dec_action_s *tail;
};

/* the bindmask is the mask of buttons that are direct bound to keycodes */
unsigned int					bindmask = 0;
struct dec_action_s	bindaction[TR_LAST];

struct dec_actiontable_s *	def_actions = NULL;
struct dec_actiontable_s *	cur_actions = NULL;

#ifndef LBUFSIZE
#define LBUFSIZE	2048
#endif /* ifndef LBUFSIZE */

int
decode_buttonbyname(char *name)
{
	if (!strcmp("ALT", name)) {
		return TR_ALT;
	} else if (!strcmp("FN", name)) {
		return TR_FN;
	} else if (!strcmp("CTRL", name)) {
		return TR_CTRL;
	} else if (!strcmp("NUM", name)) {
		return TR_NUM;
	} else if (!strcmp("MOUSE", name)) {
		return TR_MOUSE;
	} else if (!strcmp("SHIFT", name)) {
		return TR_SHIFT;
	}
	return -1;
}

void
decode_loadmap(char *mapfile)
{
	FILE *		fh = NULL;
	char		linebuffer[LBUFSIZE];
	struct dec_actiontable_s	*target = NULL;
	
	fh = fopen(mapfile, "rt");
	if (NULL == fh) {
		perror("decode_loadmap");
		exit(1);
	}
	
	while (NULL != fgets(linebuffer, LBUFSIZE, fh)) {
		char	*t;
		
		if (feof(fh))
			break;
		if (ferror(fh))
			break;
		
		t= strtok(linebuffer, CONFDELIM);
		if (NULL == t) {
			continue;
		}
		if (!strcmp("BIND", t)) {
			int		btncode = 0;
			
			if (target != NULL) {
				fprintf(stderr, "Error parsing map: Can not BIND once a keytable has been started.\n");
				exit (1);
			}
			/* bind a button */
			t = strtok(NULL, CONFDELIM);
			btncode = decode_buttonbyname(t);
			if (btncode < 0) {
				fprintf(stderr, "Error parsing map: unknown button \"%s\".\n", t);
				exit (1);
			}
			if (bindmask & (1<<btncode)) {
				fprintf(stderr, "Error parsing map: button %s is already bound.\n", t);
				exit (1);
			}
			bindmask |= (1<<btncode);
			/* reset bind action */
			memset(&(bindaction[btncode]), 0, sizeof(*bindaction));
			bindaction[btncode].next = NULL;
			t = strtok(NULL, CONFDELIM);
			/* decode action type */
			if (!strcmp("KEY", t)) {
				int		newscan;
				
				bindaction[btncode].action = ACT_SCANCODE;
				t = strtok(NULL, CONFDELIM);
				newscan = km_getscancode(t);
				if (newscan < 0) {
					fprintf(stderr, "Error parsing map: unknown key \"%s\".\n", t);
					exit (1);
				}
				bindaction[btncode].scancode = (unsigned short)newscan;
			} else if (!strcmp("MOUSE", t)) {
				bindaction[btncode].action = ACT_MOUSE;	
			} else {
				fprintf(stderr, "Error parsing map: unsupported action \"%s\" for bind.\n", t);
				exit (1);
			}
		} else if (!strcmp("DEFAULTMAP", t)) {
			/* bind to the default keyboard map */
			target = def_actions;
		} else if (!strcmp("MAP", t)) {
			struct dec_action_s	*newact;
			int		i;
			/* add a key mapping to the map */
			if (NULL == target) {
				fprintf(stderr, "Error parsing map: MAP directive outside of key map.\n");
				exit (1);
			}
			newact = malloc(sizeof(*newact));
			memset(newact, 0, sizeof(*newact));
			
			t = strtok(NULL, CONFDELIM); i = 0;
			if (strlen(t) == 4) {
				for (i = 0; i < 4; i++) {
					switch (t[i]) {
					case '-':
					case '0':
						newact->keystate[i] = TB_NONE;
						break;
					case 'L':
						newact->keystate[i] = TB_LEFT;
						break;
					case 'M':
						newact->keystate[i] = TB_MIDDLE;
						break;
					case 'R':
						newact->keystate[i] = TB_RIGHT;
						break;
					default:
						i=10; /* HACK! */
						break;	
					}
				}
				if (i >= 10) {
					i = 0;
				} else {
					i = 1;
				}
							
			}
			if (!i) {
				fprintf(stderr,"Error parsing map: Do not understand \"%s\" as key state.\n", t);
				exit (1);
			}
			t = strtok(NULL, CONFDELIM);
			/* decode action type */
			if (!strcmp("KEY", t)) {
				int		newscan;
				
				newact->action = ACT_SCANCODE;
				t = strtok(NULL, CONFDELIM);
				newscan = km_getscancode(t);
				if (newscan < 0) {
					fprintf(stderr, "Error parsing map: unknown key \"%s\".\n", t);
					exit (1);
				}
				newact->scancode = (unsigned short)newscan;
			} else if (!strcmp("MACRO", t)) {
				struct dec_macro_s		*l;
				struct dec_macro_s		*n;
				newact->action = ACT_MACRO;				
				l = n = NULL;

				t = strtok(NULL, CONFDELIM);
				if (NULL == t) {
					fprintf(stderr, "Error parsing map: expected \"{\", got EOL.\n");
					exit (1);
				}
				if (strcmp("{", t)) {
					fprintf(stderr, "Error parsing map: expected \"{\", got \"%s\".\n", t);
					exit (1);
				}					
				while (1) {
					int		newscan, soffs;
					
					t = strtok(NULL, CONFDELIM);
					if (t == NULL) {
						fprintf(stderr, "Error parsing map: Line ended without ending macro.\n");
						exit (1);
					}
					if (!strcmp("}", t)) {
						break;
					}
					n = malloc(sizeof(*n));
					memset(n, 0, sizeof(*n));
					n->next = NULL;
					
					switch (t[0]) {
					case '-':
						n->action = MAC_KEYUP;
						soffs = 1;
						break;
					case '+':
						n->action = MAC_KEYDOWN;
						soffs = 1;
						break;
					default:
						n->action = MAC_SCANCODE;
						soffs = 0;
						break;
					}
					newscan = km_getscancode(t+soffs);
					if (newscan < 0) {
						fprintf(stderr, "Error parsing map: did not recognise keyname \"%s\".\n", t+soffs);
						exit (1);
					}
					n->data = (unsigned short) newscan;
					if (NULL == l) {
						newact->macrostart = n;
					} else {
						l->next = n;
					}
					l = n;
				}
			} else {
				fprintf(stderr, "Error parsing map: unsupported action \"%s\" for map.\n", t);
				exit (1);
			}
			if (target->tail == NULL) {
				target->first = newact;
			} else {
				target->tail->next = newact;
			}	
			target->tail = newact;	
			newact->next = NULL;
		} else if (!strcmp("ALTMAP", t)) {
			struct dec_actiontable_s	*newatable;
			int							btncode;
			
			/* perform a bind to a new keyboard map */
			t = strtok(NULL, CONFDELIM);
			btncode = decode_buttonbyname(t);
			if (btncode < 0) {
				fprintf(stderr, "Error parsing map: unknown button \"%s\".\n", t);
				exit (1);
			}
			if (bindmask & (1<<btncode)) {
				fprintf(stderr, "Error parsing map: button %s is already bound.\n", t);
				exit (1);
			}
			bindmask |= (1<<btncode);
			/* reset bind action */
			memset(&(bindaction[btncode]), 0, sizeof(*bindaction));
			bindaction[btncode].next = NULL;

			bindaction[btncode].action = ACT_ALTTABLE;
			
			newatable = malloc(sizeof(*newatable));
			memset(newatable, 0, sizeof(*newatable));
			newatable->first = NULL;
			newatable->tail = NULL;

			bindaction[btncode].newtable = newatable;
			target = newatable;		
		}
	}
	
}

void
decode_init(char *mapfile)
{
	int		ctr;
	
	for (ctr = 0; ctr < TR_LAST; ctr++) {
		memset(&bindaction[ctr], 0, sizeof(bindaction[ctr]));
	}
	def_actions = malloc(sizeof(*def_actions));
	if (NULL == def_actions) {
		perror("decode_init");
		exit (1);
	}
	memset(def_actions, 0, sizeof(*def_actions));
	def_actions->first = NULL;
	def_actions->tail = NULL;
	
	decode_loadmap(mapfile);
	
	cur_actions = def_actions;
}

static void 
decode_keydown(enum thw_keyrow_e row, int button)
{
	if (bindmask & (1<<row)) {
		/* this is a 'bind' button. */
		if (bindaction[row].action == ACT_SCANCODE) {
			odev_sendkeydown(bindaction[row].scancode);	
		} else if (bindaction[row].action == ACT_ALTTABLE) {
			cur_actions = bindaction[row].newtable;
		}
		return;
	}
    keystate[row] = button;
    emitkey = 1;
}

static void 
decode_keyup(enum thw_keyrow_e row, int button)
{
	if (bindmask & (1<<row)) {
		/* this is a 'bind' button. */
		if (bindaction[row].action == ACT_SCANCODE) {
			odev_sendkeyup(bindaction[row].scancode);	
		} else if (bindaction[row].action == ACT_ALTTABLE) {
			cur_actions = def_actions;
		}
		return;
	} else {
		struct dec_action_s		*p = cur_actions->first;

		if (emitkey) {
			emitkey = 0;
			while (p != NULL) {
				int		i = 0;
				int		m = 0;
				for (i = 0; i < 4; i++) {
					if (keystate[TR_ROW1+i] == p->keystate[i]) {
						m++;
					}
				}
				if (4 == m) {
					/* match! send execute action and return */	
					if (p->action == ACT_MACRO) {
						struct dec_macro_s		*mac = p->macrostart;
						
						while (mac != NULL) {
							switch (mac->action) {
							case MAC_SCANCODE:
								odev_sendkeydown(mac->data);
							case MAC_KEYUP:
								odev_sendkeyup(mac->data);
								break;
							case MAC_KEYDOWN:
								odev_sendkeydown(mac->data);
								break;
							default:
								break;
							}
							mac = mac->next;
						}
					} else if (p->action == ACT_SCANCODE) {
						odev_sendkeydown(p->scancode);
						odev_sendkeyup(p->scancode);
					}
					/* break out */
					goto finish;
				}
				p = p->next;
			}
		}
finish:
		keystate[row] = 0;
	}
}

struct thw_callback_s decode_callback = {
  decode_keydown,
  decode_keyup,
  NULL
};
