/* twidhw.c  ****************************************************************
 
 Twiddler Driver (userspace)
 -- Twiddler1 Hardware handling
 
 Copyright (C) 2005, Chris Collins <chris@collins.id.au>
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author, Christopher Collins, nor any of his
    aliases may be used to endorse or promote products derived
    from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY CHRISTOPHER COLLINS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN NO EVENT SHALL CHRISTOPHER COLLINS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <termios.h>

#include "twiddled.h"

static int		fd = -1;

int
thw_open(char *devpath)
{
  int			line;
  struct termios	tca;

  atexit(thw_close);
  fd = open(devpath, O_RDONLY);
  if (fd < 0){
    perror("Couldn't open input device");
    exit (1);
  }

  /* set device to twiddler parameters */
  if (tcgetattr(fd, &tca)) {
    perror("Getting serial port attributes");
    exit (1);
  }
  /* put in raw data mode */
  tca.c_lflag &= ~(ICANON | ECHO | IEXTEN | ISIG);
  tca.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  tca.c_cflag &= ~(CSIZE | PARENB);
  tca.c_oflag &= ~OPOST;
  tca.c_cflag |= CS8; /* set to 8 stop bits */
  tca.c_cflag |= CLOCAL;
  tca.c_cc[VMIN]=1;
  tca.c_cc[VTIME]=0;
  cfsetospeed (&tca, B2400);
  cfsetispeed (&tca, B2400);
  if (tcsetattr(fd, TCSAFLUSH, &tca)) {
    perror("Setting serial port attributes");
    exit (1);
  }
  tcflush (fd, TCIOFLUSH);
  tcflow (fd,TCOOFF);
  ioctl (fd, TIOCMGET, &line);
  line &= ~TIOCM_DTR;
  ioctl (fd, TIOCMSET, &line);

  return fd;
}

void
thw_close()
{
  if (fd > 0) {
    close (fd);
    fd = -1;
  }
}



static unsigned int	last_facekeys = 0;
static unsigned int	last_thumbkeys = 0;

static int	last_xpos = 0;
static int	last_ypos = 0;

#ifndef BUFSIZE
#define BUFSIZE 16
#endif

void
thw_update(struct thw_callback_s *callbacks)
{
  int	syncerror = 0;
  unsigned char	buffer[BUFSIZE];
  int	bufoffs = 0;
  unsigned int	facekeys = 0;
  unsigned int	thumbkeys = 0;
  int	xpos = 0;
  int	ypos = 0;
  int	delta;
 
  /* pull in 5 characters (a twiddler packet) */
  while (bufoffs < 5) {
    bufoffs += read(fd, buffer+bufoffs, 5-bufoffs);
  }

  /* resynchronise if necessary */
  for (syncerror = 0; buffer[syncerror] & 0x80; syncerror++)
    ;
  if (syncerror) {
    while ((bufoffs-5) < syncerror) {
      bufoffs += read(fd, buffer+bufoffs, syncerror - (bufoffs-5));
    }
  }

  /* unpack! */
  facekeys = ((buffer[syncerror+1]&0x1) << 7) 
    | (buffer[syncerror] & 0x7F);
  thumbkeys = (buffer[syncerror+1]&0x7E) >> 1;


  ypos = (buffer[syncerror+2]&0x7f) |
    ((buffer[syncerror+3]&0x3)<<7);

  xpos = ((buffer[syncerror+3]&0x7c)>>2) |
    ((buffer[syncerror+4]&0xf)<<5);

  /* first up, report movement */
  if (callbacks->relmove != NULL) {
    if (xpos != last_xpos || ypos != last_ypos) {
      callbacks->relmove(xpos-last_xpos, ypos-last_ypos);
    }
  }
  last_ypos = ypos;
  last_xpos = xpos;

  /* check thumb butons for releases */
  delta = thumbkeys ^ last_thumbkeys;
  if (delta && callbacks->keyup != NULL) {
    if ((delta & 0x1) && (last_thumbkeys & 0x1)) {
      /* shift */
      callbacks->keyup(TR_SHIFT, 1);
    }
    if ((delta & 0x2) && (last_thumbkeys & 0x2)) {
      /* num */
      callbacks->keyup(TR_NUM, 1);
    }
    if ((delta & 0x4) && (last_thumbkeys & 0x4)) {
      /* fn */
      callbacks->keyup(TR_FN, 1);
    }
    if ((delta & 0x8) && (last_thumbkeys & 0x8)) {
      /* ctrl */
      callbacks->keyup(TR_CTRL, 1);
    }
    if ((delta & 0x10) && (last_thumbkeys & 0x10)) {
      /* alt */
      callbacks->keyup(TR_ALT, 1);
    }
    if ((delta & 0x20) && (last_thumbkeys & 0x20)) {
      /* mouse */
      callbacks->keyup(TR_MOUSE, 1);
    }
  }

  /* next, check for face key movement */
  delta = facekeys ^ last_facekeys;
  if (delta && (callbacks->keyup != NULL)) {
    int		ctr;
    /* do key releases */
    for (ctr = 0; ctr < 4; ctr++) {
      if ((delta>>(ctr*2)) & 0x3) {
	/* key has changed... */
	int okval = (last_facekeys >> (ctr*2))&0x3;

	if (okval) {
	  callbacks->keyup(ctr+1, okval);
	}
      }
    }
  }

  /* check thumb butons for presses */
  delta = thumbkeys ^ last_thumbkeys;
  if (delta && callbacks->keydown != NULL) {
    if ((delta & 0x1) && (thumbkeys & 0x1)) {
      /* shift */
      callbacks->keydown(TR_SHIFT, 1);
    }
    if ((delta & 0x2) && (thumbkeys & 0x2)) {
      /* num */
      callbacks->keydown(TR_NUM, 1);
    }
    if ((delta & 0x4) && (thumbkeys & 0x4)) {
      /* fn */
      callbacks->keydown(TR_FN, 1);
    }
    if ((delta & 0x8) && (thumbkeys & 0x8)) {
      /* ctrl */
      callbacks->keydown(TR_CTRL, 1);
    }
    if ((delta & 0x10) && (thumbkeys & 0x10)) {
      /* alt */
      callbacks->keydown(TR_ALT, 1);
    }
    if ((delta & 0x20) && (thumbkeys & 0x20)) {
      /* mouse */
      callbacks->keydown(TR_MOUSE, 1);
    }
  }

  /* do key presses */
  delta = facekeys ^ last_facekeys;
  if (delta && (callbacks->keydown != NULL)) {
    int		ctr;

    for (ctr = 0; ctr < 4; ctr++) {
      if ((delta>>(ctr*2)) & 0x3) {
	/* key has changed... */
	int kval = (facekeys >> (ctr*2))&0x3;

	if (kval) {
	    callbacks->keydown(ctr+1, kval);
	}
      }
    }    
  }

  /* update button states */
  last_facekeys = facekeys;
  last_thumbkeys = thumbkeys;

  return;
}
